r"""This package contains the model base class and some implemented
models."""

from __future__ import annotations

__all__ = [
    "BaseModel",
    "VanillaModel",
    "attach_module_to_engine",
    "is_model_config",
    "setup_and_attach_model",
    "setup_model",
]

from gravitorch.models.base import BaseModel
from gravitorch.models.utils import (
    attach_module_to_engine,
    is_model_config,
    setup_and_attach_model,
    setup_model,
)
from gravitorch.models.vanilla import VanillaModel
